/* source: xio-dccp.h */
/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
/* Published under the GNU General Public License V.2, see file COPYING */

#ifndef __xio_dccp_h_included
#define __xio_dccp_h_included 1

extern const struct addrdesc xioaddr_dccp_connect;
extern const struct addrdesc xioaddr_dccp_listen;
extern const struct addrdesc xioaddr_dccp4_connect;
extern const struct addrdesc xioaddr_dccp4_listen;
extern const struct addrdesc xioaddr_dccp6_connect;
extern const struct addrdesc xioaddr_dccp6_listen;

extern const struct optdesc xioopt_dccp_set_ccid;

#endif /* !defined(__xio_dccp_h_included) */
