/* source: xio-tun.h */
/* Copyright Gerhard Rieger and contributors (see file CHANGES) */
/* Published under the GNU General Public License V.2, see file COPYING */

#ifndef __xio_tun_h_included
#define __xio_tun_h_included 1

extern const struct optdesc opt_tun_device;
extern const struct optdesc opt_tun_name;
extern const struct optdesc opt_tun_type;
extern const struct optdesc opt_iff_no_pi;

extern const struct addrdesc xioaddr_tun;

#endif /* !defined(__xio_tun_h_included) */
